package flowserver

import (
	"encoding/json"
	"errors"
	"flow"
	"flow/registry"
	"fmt"
	"log"
	"strconv"
	"time"
)

// Node that will contain registry src
type Node struct {
	ID            string            `json:"id"`
	Src           string            `json:"src"`
	Label         string            `json:"label"`
	DefaultInputs map[int]string    `json:"defaultInputs"`
	Prop          map[string]string `json:"prop"`
}

// Link that joins two nodes
type Link struct {
	From string `json:"from"`
	To   string `json:"to"`
	In   int    `json:"in"`
}

// Trigger that join two nodes on state change
type Trigger struct {
	From string   `json:"from"`
	To   string   `json:"to"`
	On   []string `json:"on"`
}

// FlowDocument flow document
type FlowDocument struct {
	Nodes    []Node    `json:"nodes"`
	Links    []Link    `json:"links"`
	Triggers []Trigger `json:"triggers"`
}

func (fd *FlowDocument) fetchNodeByID(ID string) *Node {
	for _, n := range fd.Nodes {
		if n.ID == ID {
			return &n
		}
	}
	return nil
}

func (fd *FlowDocument) fetchTriggerFrom(ID string) []Trigger {
	ret := []Trigger{}
	for _, t := range fd.Triggers {
		if t.From != ID {
			continue
		}
		log.Println("Trigger is:", ID, "adding")
		ret = append(ret, t)
	}
	return ret
}

func (fd *FlowDocument) fetchLinksTo(ID string) []Link {
	ret := []Link{}
	for _, l := range fd.Links {
		if l.To != ID {
			continue
		}
		ret = append(ret, l)
	}
	return ret
}

var ErrLoop = errors.New("Looping through is disabled for now")

// FlowBuild build the graph based on an starting node
func FlowBuild(rawData []byte, r *registry.R, startingID string) (*flow.Flow, error) {

	doc := FlowDocument{[]Node{}, []Link{}, []Trigger{}}
	err := json.Unmarshal(rawData, &doc)
	if err != nil {
		return nil, err
	}

	f := flow.New()
	f.SetRegistry(r)

	nodeTrack := map[string]bool{}

	var build func(ID string) (flow.Operation, error)
	build = func(ID string) (flow.Operation, error) {
		if _, ok := nodeTrack[ID]; ok {
			return nil, ErrLoop //fmt.Errorf("[%v] Looping through nodes is disabled:", ID)
		}
		nodeTrack[ID] = true

		// If flow already has ID just return
		if op := f.GetOp(ID); op != nil {
			log.Println("Return operation")
			return op, nil
		}

		node := doc.fetchNodeByID(ID)
		if node == nil {
			return nil, fmt.Errorf("node not found [%v]", startingID)
		}

		var op flow.Operation

		switch node.Src {
		case "Input":
			inputID, err := strconv.Atoi(node.Prop["input"])
			if err != nil {
				return nil, errors.New("Invalid inputID value, must be a number")
			}
			op = f.In(inputID) // By id perhaps
		case "Variable":
			op = f.Var(node.ID, node.Prop["init"])
		case "Const":
			var val flow.Data
			// XXX: Automate this in a func
			raw := node.Label
			err := json.Unmarshal([]byte(raw), &val)
			if err != nil { // Try to unmarshal as a string?
				val = string(raw)
			}
			op, _ = f.DefConst(node.ID, val)
		default:
			// Load entry
			entry, err := r.Entry(node.Src)
			if err != nil {
				return nil, err
			}
			inputs := doc.fetchLinksTo(node.ID)
			param := make([]flow.Data, len(entry.Inputs))
			for _, l := range inputs {
				param[l.In], err = build(l.From)
				if err != nil {
					return nil, err
				}
				// if type Different than ours try to convert
			}
			op, err = f.DefOp(node.ID, node.Src, param...)
			if err != nil {
				return nil, err
			}
		}

		// Process triggers for this node
		triggers := doc.fetchTriggerFrom(node.ID)
		for _, t := range triggers {
			_, err := build(t.To)
			if err != nil {
				return nil, err
			}
			// Register the thing here
			f.Hook(flow.Hook{
				Any: func(name string, ID string, triggerTime time.Time, extra ...interface{}) {
					if name != "Error" && name != "Finish" {
						return
					}
					if ID != t.From {
						log.Printf("ID[%v] triggered [%v], I'm t.From: %v", ID, name, t.From)
						return
					}
					exec := false
					for _, o := range t.On {
						if name == o {
							exec = true
							break
						}
					}
					if !exec {
						log.Println("Mismatching trigger, but its a test")
					}

					op := f.Res(t.To)   // Repeating
					go op.Process(name) // Background
				},
			})

			delete(nodeTrack, node.ID)
		}
		return op, nil
	}

	_, err = build(startingID)
	if err != nil {
		log.Println("Error building")
		return nil, err
	}

	return f, nil
}
